@@ -1,4 +1,4 @@
-#!/usr/bin/python
+#!/usr/bin/python -O
 
 """collect hits per ASN and netblock"""
 
@@ -30,14 +30,16 @@ import json
 import logging
 import os
 import re
+import shlex
 import sys
 import threading
 from collections import Counter
 from functools import partial
+from io import StringIO
 from pathlib import Path
 from signal import SIGHUP, signal
 from subprocess import CalledProcessError, run
-from typing import Any, Callable, TextIO
+from typing import Any, Callable, Iterator, TextIO
 
 try:
     import netaddr
@@ -49,6 +51,8 @@ except ImportError:
 
 
 try:
+    from wsgiref.simple_server import WSGIServer
+
     import prometheus_client
 
     prometheus_client_available = True
@@ -84,7 +88,7 @@ TCPDUMP_REGEX = re.compile(
 [ ]IP6?[ ]
 (?:\[[^]]+][ ]\(invalid\)[ ])?  # anomalies like "[total length 52 > length 44] (invalid)"
 (?P<src>[0-9a-fA-F:.]+)  # source, e.g. 203.0.113.42.62122
-[ ]>[ ]  # seperator, e.g. " > "
+[ ]>[ ]  # separator, e.g. " > "
 (?P<dst>[0-9a-fA-F:.]+)  # destination, e.g. 2001:DB8::2.443
 :.* # whatever else follows
 """,
@@ -96,16 +100,26 @@ class Recorder:
     asndb = None
     asnamesdb = None
 
-    def record(self, asn: int | None, prefix: str | None) -> None:
+    def record(
+        self, asn: int | None, prefix: str | None, count: float | int = 1
+    ) -> None:
         raise NotImplementedError()
 
     def display_results(self, stream: TextIO = sys.stdout) -> None:
         raise NotImplementedError()
 
-    def _load_asndb(self):
+    def shutdown(self) -> None:
+        """garbage collection routines"""
+        pass
+
+    def _load_asndb(self) -> None:
         """ensure asn database is loaded"""
         if self.asndb is None:
-            datfile = str(find_cachefiles(DATFILE_GLOB)[-1].absolute())
+            datfiles = find_cachefiles(DATFILE_GLOB)
+            assert (
+                datfiles
+            ), "could not find cache file, download failed and error uncaught?"
+            datfile = str(datfiles[-1].absolute())
             logging.info("loading datfile %s...", datfile)
             self.asndb = pyasn.pyasn(datfile)
 
@@ -116,7 +130,8 @@ class Recorder:
         """
         self._load_asndb()
         assert self.asndb
-        return self.asndb.lookup(address)
+        # upstream lacks type signatures
+        return self.asndb.lookup(address)  # type: ignore[no-any-return]
 
     def lookup_asn(self, asn: int) -> str | None:
         """lookup the AS name based on the number
@@ -128,12 +143,20 @@ class Recorder:
             if not asfile.exists():
                 download_asnames()
             logging.info("loading %s", str(asfile))
-            with asfile.open() as fp:
-                self.asnamesdb = json.load(fp)
+            try:
+                with asfile.open() as fp:
+                    self.asnamesdb = json.load(fp)
+            except OSError as e:
+                logging.warning(
+                    "could not load asnames.json, disabling AS names lookups: %s", e
+                )
+                args.no_resolve_asn = True
+                return None
 
-        return self.asnamesdb.get(str(asn))
+        # assert that we get a string back from the loader
+        return self.asnamesdb.get(str(asn))  # type: ignore[no-any-return]
 
-    def asn_all_prefixes(self, asn: int, aggregate: bool = False) -> list[str]:
+    def asn_all_prefixes(self, asn: int, aggregate: bool = False) -> set[str]:
         """all prefixes from the given ASN
 
         This is *all* prefixes from the routing table, not just the
@@ -141,12 +164,13 @@ class Recorder:
         """
         self._load_asndb()
         assert self.asndb
+        if aggregate and not netaddr_available:
+            logging.warning("netaddr not available, cannot aggregate results")
+            aggregate = False
         if not aggregate:
-            return self.asndb.get_as_prefixes(asn)
-        if netaddr_available:
-            logging.error("netaddr not available, cannot aggregate results")
-            return []
-        return netaddr.cidr_merge(self.asndb.get_as_prefixes(asn))
+            # upstream lacks type signatures
+            return self.asndb.get_as_prefixes(asn)  # type: ignore[no-any-return]
+        return {str(n) for n in netaddr.cidr_merge(self.asndb.get_as_prefixes(asn))}
 
     def asn_all_prefixes_str(
         self, asn: int, aggregate: bool = False, sep: str = "\n"
@@ -160,14 +184,31 @@ class Recorder:
         """
         return sep.join(sorted(self.asn_all_prefixes(asn, aggregate)))
 
+    def asn_prefixes(self, *asn: int, aggregate: bool = False) -> set[str]:
+        raise NotImplementedError()
+
+    def asn_prefixes_str(
+        self, *asn: int, aggregate: bool = False, sep: str = "\n"
+    ) -> str:
+        """all prefixes found matching the given ASN
+
+        This is *not* all the known prefixes for this ASN, only the
+        ones we matched on. See asn_all_prefixes for the latter.
+
+        Same as asn_prefixes, but joined on a the given sep (defaults newline).
+        """
+        return sep.join(sorted(self.asn_prefixes(*asn, aggregate)))
+
 
 class CollectionsRecorder(Recorder):
     def __init__(self) -> None:
-        self.asn_counter: Counter = Counter()
+        self.asn_counter: Counter[int | None] = Counter()
         # optimise with pyradix, if necessary
-        self.prefix_counter: Counter = Counter()
+        self.prefix_counter: Counter[str | None] = Counter()
 
-    def record(self, asn: int | None, prefix: str | None) -> None:
+    def record(
+        self, asn: int | None, prefix: str | None, count: int | float = 1
+    ) -> None:
         """record a hit for the match in regular collection counters
 
         This *could* be optimized a little further, by moving the "if"s
@@ -175,10 +216,10 @@ class CollectionsRecorder(Recorder):
         """
         # record the ASN, if relevant
         if not args.no_asn:
-            self.asn_counter[asn] += 1
+            self.asn_counter[asn] += round(count)
         # record the prefix, if relevant
         if not args.no_prefixes:
-            self.prefix_counter[prefix] += 1
+            self.prefix_counter[prefix] += round(count)
         # we resolve ASNs later, on display, in this pattern
 
     def display_results(self, stream: TextIO = sys.stdout) -> None:
@@ -218,11 +259,12 @@ class CollectionsRecorder(Recorder):
             ratio = 0.0
             print("count\tpercent\tprefix\tASN\tAS", file=stream)
             for prefix, count in self.prefix_counter.most_common(args.top):
-                asn = asname = ""
+                asn = None
+                asname = ""
                 if prefix:
                     address, _ = prefix.split("/")
                     asn, _ = self.lookup_address(address)
-                    if not args.no_resolve_asn:
+                    if asn and not args.no_resolve_asn:
                         asname = self.lookup_asn(asn) or ""
                 if total:
                     ratio = round(count / total * 100, 2)
@@ -242,37 +284,36 @@ class CollectionsRecorder(Recorder):
             for prefix in self.asndb.get_as_prefixes(a) or []:
                 if prefix in self.prefix_counter.keys():
                     prefixes.add(prefix)
+        if aggregate and not netaddr_available:
+            aggregate = False
+            logging.warning("netaddr not available, cannot aggregate results")
         if not aggregate:
             return prefixes
-        if netaddr_available:
-            logging.error("netaddr not available, cannot aggregate results")
-            return {}
-        return netaddr.cidr_merge(prefixes)
-
-    def asn_prefixes_str(
-        self, *asn: int, aggregate: bool = False, sep: str = "\n"
-    ) -> str:
-        """all prefixes found matching the given ASN
-
-        This is *not* all the known prefixes for this ASN, only the
-        ones we matched on. See asn_all_prefixes for the latter.
-
-        Same as asn_prefixes, but joined on a the given sep (defaults newline).
-        """
-        return sep.join(sorted(self.asn_prefixes(*asn, aggregate)))
+        return {str(n) for n in netaddr.cidr_merge(prefixes)}
 
 
 class PrometheusRecorder(Recorder):
 
-    def __init__(self) -> None:
+    def __init__(self, port: int | None) -> None:
         self.asn_metrics_counter = prometheus_client.Counter(
             "asncounter_asn", "hits per AS number", ["asn", "name"]
         )
         self.prefix_metrics_counter = prometheus_client.Counter(
             "asncounter_prefix", "hits per CIDR prefix", ["cidr", "asn", "name"]
         )
+        self.prom_server: WSGIServer | None = None
+        self.prom_thread: threading.Thread | None = None
+        if port:
+            try:
+                self.prom_server, self.prom_thread = (
+                    prometheus_client.start_http_server(args.port)
+                )
+            except TypeError:
+                # older prometheus_client libraries don't return the server object
+                pass
+            logging.info("Prometheus metrics exposed on port %s", port)
 
-    def record(self, asn: int | None, prefix: str | None) -> None:
+    def record(self, asn: int | None, prefix: str | None, count: float = 1.0) -> None:
         """record a hit for the match in Prometheus
 
         This probably needs to be optimized, as it's inside a hot loop and
@@ -292,13 +333,13 @@ class PrometheusRecorder(Recorder):
 
         # record the ASN, if relevant
         if not args.no_asn:
-            self.asn_metrics_counter.labels(asn, name).inc()
+            self.asn_metrics_counter.labels(asn, name).inc(count)
 
         # record the prefix, if relevant
         if not args.no_prefixes:
             self.prefix_metrics_counter.labels(
                 prefix, asn if not args.no_asn else "", name
-            ).inc()
+            ).inc(count)
 
     def display_results(self, stream: TextIO = sys.stdout) -> None:
         """display results from the Prometheus metrics
@@ -326,6 +367,12 @@ class PrometheusRecorder(Recorder):
         output = prometheus_client.generate_latest(registry).decode("utf-8")  # type: ignore
         print(output, file=stream, end="")
 
+    def shutdown(self) -> None:
+        if self.prom_server:
+            assert self.prom_thread
+            self.prom_server.shutdown()
+            self.prom_thread.join()
+
 
 class Collector:
     def __init__(self, recorder: Recorder) -> None:
@@ -333,52 +380,47 @@ class Collector:
 
     def _collect(
         self,
-        stream: TextIO = sys.stdin,
-        parser: Callable[[str], str | None] | None = None,
+        stream: Iterator[str],
     ) -> None:
         raise NotImplementedError()
 
-    def collect(
-        self,
-        stream: TextIO = sys.stdin,
-        parser: Callable[[str], str | None] | None = None,
-    ) -> None:
+    def collect(self, stream: Iterator[str]) -> None:
+        logging.info(
+            "collecting addresses in %s mode",
+            type(self).__name__,
+        )
         try:
-            self._collect(stream, parser)
+            self._collect(stream)
         except KeyboardInterrupt:
             pass
         logging.info("finished reading data")
 
+    def parse(self, address: str) -> tuple[str | None, float]:
+        raise NotImplementedError()
+
 
 class ScapyCollector(Collector):
     def _collect(
         self,
-        stream: TextIO = sys.stdin,
-        parser: Callable[[Packet], str | None] | None = None,  # type: ignore[override]
+        stream: Iterator[str],
     ) -> None:
-        if parser is None:
-            parser = self._parse
         if args.interface:
             logging.info("reading packets from interface %s", args.interface)
             sniff(
                 filter=args.scapy_filter,
-                prn=parser,
+                prn=self._parse,
                 store=False,
                 iface=args.interface,
             )
         else:
-            logging.info("reading packets stream %s", getattr(stream, "name", stream))
             sniff(
                 filter=args.scapy_filter,
-                prn=parser,
+                prn=self._parse,
                 store=False,
                 offline=args.input,
             )
 
-    def _parse(
-        self,
-        packet: Packet,
-    ) -> None:
+    def _parse(self, packet: Packet) -> None:
         """callback for the scapy sniffer
 
         Essentially just lookup the packet's IP address and do the rest of
@@ -395,35 +437,32 @@ class ScapyCollector(Collector):
                 packet.summary(),
             )
             return
-        # too hot, about 1ms / 1k packets
-        # logging.debug("IP %s matched to ASN %s prefix %s", address, asn, prefx)
+        # this is 3 object lookups, a function call, and a dict lookup
+        if __debug__:
+            # too hot to run systematically, about 1ms / 1k packets
+            logging.debug("IP %s matched to ASN %s prefix %s", address, asn, prefix)
         self.recorder.record(asn, prefix)
 
 
-class LineCollector(Collector):
-    def _collect(
-        self,
-        stream: TextIO = sys.stdin,
-        parser: Callable[[str], str | None] | None = None,
-    ) -> None:
+class BaseLineCollector(Collector):
+    def _collect(self, stream: Iterator[str]) -> None:
         """record IP addresses from stream, with the given parser
 
         By default, the parser is a noop, that is each line is an IP
         address, but parse_tcpdump can be used to extract the source IP
         address.
         """
-        if parser is None:
-
-            def parser(address: str | None = None):
-                return address
-
-        logging.info("collecting addresses from %s", getattr(stream, "name", stream))
         for line in stream:
-            line = line.strip()
-            # skip empty lines and lines starting with a comment
-            if not line or line[0] == "#":
+            # strip trailing comments and whitespace
+            line = line.split("#", maxsplit=1)[0].strip()
+            # skip empty lines
+            if not line:
                 continue
-            address = parser(line)
+            address: str | None
+            if self.parse is None:
+                address, count = line, 1.0
+            else:
+                address, count = self.parse(line)
             if not address:
                 logging.warning("failed to parse line, skipped: %s", line)
                 continue
@@ -438,40 +477,55 @@ class LineCollector(Collector):
                 )
                 continue
             # too hot, about 1ms / 1k packets
-            # logging.debug("IP %s matched to ASN %s prefix %s", address, asn, prefx)
-            self.recorder.record(asn, prefix)
+            if __debug__:
+                logging.debug("IP %s matched to ASN %s prefix %s", address, asn, prefix)
+            self.recorder.record(asn, prefix, count)
 
 
-class TcpdumpLineCollector(LineCollector):
-    def _collect(
-        self,
-        stream: TextIO = sys.stdin,
-        parser: Callable[[str], str | None] | None = None,
-    ) -> None:
-        """record IP addresses from stream, with the tcpdump parser
+class LineCollector(BaseLineCollector):
+    parse = None  # type: ignore[assignment]
 
-        By default, the parser is a noop, that is each line is an IP
-        address, but parse_tcpdump can be used to extract the source IP
-        address.
-        """
-        if parser is None:
-            parser = self.parse
-        super()._collect(stream, parser)
 
-    def parse(self, line: str) -> str | None:
+class TupleCollector(BaseLineCollector):
+    """record IP addresses from stream, with the given parser
+
+    By default, the parser is a noop, that is each line is an IP
+    address, but parse_tcpdump can be used to extract the source IP
+    address.
+    """
+
+    def parse(self, line: str) -> tuple[str | None, float]:
+        if not line:
+            return None, 0.0
+        try:
+            address, count = line.split(maxsplit=1)
+        except ValueError:
+            return None, 0.0
+        return address, float(count)
+
+
+class TcpdumpLineCollector(BaseLineCollector):
+    """record IP addresses from stream, with the tcpdump parser
+
+    By default, the parser is a noop, that is each line is an IP
+    address, but parse_tcpdump can be used to extract the source IP
+    address.
+    """
+
+    def parse(self, line: str) -> tuple[str | None, float]:
         """parse a tcpdump line for source/destination IP addresses"""
         m = TCPDUMP_REGEX.match(line)
         if not m:
             logging.warning("cannot parse tcpdump line %s, ignoring", line.strip())
-            return None
+            return None, 0.0
 
         # we split out the port at the end, because it's too complicated
         # to do that in the regex directly, especially in IPv4 mode
         # because the separator is the same
         for address_port in (m.group("src"), m.group("dst")):
             split = address_port.rsplit(".", maxsplit=1)
-            return split[0]
-        return None
+            return split[0], 1.0
+        return None, 0.0
 
 
 def find_datfile() -> str:
@@ -502,7 +556,7 @@ def find_datfile() -> str:
 
 
 def refresh_datfile() -> str:
-    """download a ribfile and convert into a datfile unconditionnally"""
+    """download a ribfile and convert into a datfile unconditionally"""
     try:
         ribfile = download_ribfile()
         return convert_ribfile(ribfile)
@@ -535,31 +589,31 @@ def download_ribfile() -> str:
     This raises CalledProcessError, or assertion.
     """
     cmd = ["pyasn_util_download.py", "--latestv46"]
-    logging.info("downloading rib file with: %s", " ".join(cmd))
+    logging.info("downloading rib file with: %s", shlex.join(cmd))
     run(cmd, cwd=args.cache_directory, check=True)
     ribfiles = find_cachefiles(RIBFILE_GLOB)
     assert ribfiles, "no rib file found"
     return ribfiles[-1].name
 
 
-def download_asnames():
+def download_asnames() -> None:
     """download the AS mapping file in the cache directory"""
     cmd = ["pyasn_util_asnames.py", "--output", "asnames.json"]
-    logging.info("downloading AS names mappings with: %s", " ".join(cmd))
+    logging.info("downloading AS names mappings with: %s", shlex.join(cmd))
     run(cmd, cwd=args.cache_directory, check=True)
 
 
-def convert_ribfile(ribfile: str):
+def convert_ribfile(ribfile: str) -> str:
     """convert the given ribfile into a datfile"""
     datfile = guess_datfile(ribfile).removesuffix(".gz")
     cmd = ["pyasn_util_convert.py", "--compress", "--single", ribfile, datfile]
-    logging.info("converting file %s to %s with: %s", ribfile, datfile, cmd)
+    logging.info("converting file %s to %s with: %s", ribfile, datfile, shlex.join(cmd))
     # pyasn_util_convert.py --single rib.20250522.1400.bz2 ipasn_20250522.1400.dat
     run(cmd, cwd=args.cache_directory, check=True)
     return str(Path(args.cache_directory) / datfile)
 
 
-def parse_args() -> None:
+def parse_args(argv: list[str] = sys.argv[1:]) -> None:
     parser = argparse.ArgumentParser(
         description=__doc__,
         epilog=__epilog__,
@@ -568,7 +622,9 @@ def parse_args() -> None:
     parser.add_argument(
         "--cache-directory",
         "-C",
-        default=os.environ.get("XDG_CACHE_HOME", os.environ["HOME"] + "/.cache")
+        default=os.environ.get(
+            "XDG_CACHE_HOME", os.environ.get("HOME", ".") + "/.cache"
+        )
         + "/pyasn",
         help="where to store pyasn cache files, default: %(default)s",
     )
@@ -606,7 +662,7 @@ def parse_args() -> None:
         "--input-format",
         "-I",
         default="line",
-        choices=["line", "tcpdump", "scapy"],
+        choices=["line", "tuple", "tcpdump", "scapy"],
         help="input format, default: %(default)s",
     )
     parser.add_argument(
@@ -654,8 +710,40 @@ def parse_args() -> None:
         action="store_true",
         help="run a REPL thread in main loop",
     )
+    runtime_dir = os.environ.get("XDG_RUNTIME_DIR")
+    if not runtime_dir:
+        runtime_dir = os.environ.get(
+            "XDG_STATE_HOME", os.environ.get("HOME", ".") + "/.local/state"
+        )
+        print(
+            "INFO: XDG_RUNTIME_DIR not defined, falling back to",
+            runtime_dir,
+            file=sys.stderr,
+        )
+
+    parser.add_argument(
+        "--manhole",
+        const=str(Path(runtime_dir) / "asncounter-manhole-{pid}"),
+        nargs="?",
+        help="setup a REPL socket with manhole, default %(const)s if set",
+    )
     parser.add_argument("--debug", action="store_true", help="more debugging output")
-    parser.parse_args(namespace=args)
+    parser.add_argument(
+        "address", nargs="*", help="addresses to process instead of stdin"
+    )
+    parser.parse_args(argv, namespace=args)
+
+    if args.address:
+        if args.input_format != "line":
+            parser.error(
+                "cannot specify a different --input-format and command-line address: %s"
+                % shlex.join(args.address)
+            )
+        if args.input != sys.stdin:
+            parser.error(
+                "cannot provide an --input file and and command-line address: %s"
+                % shlex.join(args.address)
+            )
 
     if args.output_format == "prometheus":
         if not prometheus_client_available:
@@ -671,15 +759,58 @@ def parse_args() -> None:
         args.no_resolve_asn = True
 
 
-def main() -> None:
-    parse_args()
+def repl_thread(collector: Callable[[], None], namespace: dict[str, str]) -> None:
+    collect_thread = threading.Thread(target=collector, daemon=True)
+    collect_thread.start()
+
+    sys.stdin = open("/dev/tty")  # reopen terminal to get a fresh stdin
+    logging.info(
+        "starting interactive console, use recorder.display_results() to show current results",
+    )
+    if not args.output_format == "prometheus":
+        logging.info(
+            "recorder.asn_counter and .prefix_counter dictionaries have the full data"
+        )
+
+    repl_thread = threading.Thread(
+        target=lambda: code.interact(local=namespace), daemon=True
+    )
+    repl_thread.start()
+    try:
+        repl_thread.join()
+    except KeyboardInterrupt:
+        pass
+
+
+def manhole_thread(namespace: dict[str, str]) -> None:
+    manhole_dir = Path(args.manhole).parent
+    manhole_socket = args.manhole.format(pid=os.getpid())
+    if not manhole_dir.exists():
+        manhole_dir.mkdir(exist_ok=True, parents=True)
+    import manhole
+
+    try:
+        manhole.install(socket_path=manhole_socket, locals=namespace)
+    except Exception as e:
+        logging.warning("failed to create manhole in %s: %s", manhole_socket, e)
+
+
+def main(argv: list[str] = sys.argv[1:]) -> Recorder:
+    parse_args(argv)
 
     logging.basicConfig(
         level="DEBUG" if args.debug else "INFO", format="%(levelname)s: %(message)s"
     )
 
-    if not Path(args.cache_directory).exists():
-        Path(args.cache_directory).mkdir(parents=True)
+    if args.address:
+        logging.info(
+            "reading IP addresses from the commandline, ignoring stdin or input files"
+        )
+        args.input = StringIO("\n".join(args.address))
+    else:
+        logging.info("selected input file %s", getattr(args.input, "name", args.input))
+
+    Path(args.cache_directory).mkdir(parents=True, exist_ok=True)
 
     if args.refresh:
         refresh_datfile()
@@ -689,26 +820,20 @@ def main() -> None:
     logging.info("using datfile %s", find_datfile())
 
     # setup display procedures and prom server
-    prom_server = None
     recorder: Recorder = CollectionsRecorder()
     if args.output_format == "prometheus":
-        recorder = PrometheusRecorder()
-        if getattr(args, "port"):
-            try:
-                prom_server, prom_thread = prometheus_client.start_http_server(
-                    args.port
-                )
-            except TypeError:
-                # older prometheus_client libraries don't return the server object
-                pass
-            logging.info("Prometheus metrics exposed on port %s", args.port)
+        recorder = PrometheusRecorder(getattr(args, "port"))
+
     if args.output_format != "null":
         signal(SIGHUP, lambda s, f: recorder.display_results())
 
+    cls: type[Collector]
     if args.input_format == "tcpdump":
-        cls: type[Collector] = TcpdumpLineCollector
+        cls = TcpdumpLineCollector
     elif args.input_format == "line":
         cls = LineCollector
+    elif args.input_format == "tuple":
+        cls = TupleCollector
     elif args.input_format == "scapy":
         cls = ScapyCollector
     else:
@@ -717,42 +842,36 @@ def main() -> None:
 
     collector = partial(cls(recorder=recorder).collect, stream=args.input)
 
+    # need to extract the namespace before we enter the monitoring
+    # thread (manhole, repl), which run with different locals
+    namespace = locals() | globals()
+
+    if args.manhole:
+        manhole_thread(namespace)
+
     # main read loop
     if args.repl:
-        collect_thread = threading.Thread(target=collector, daemon=True)
-        collect_thread.start()
-
-        sys.stdin = open("/dev/tty")  # reopen terminal to get a fresh stdin
-        logging.info(
-            "starting interactive console, use recorder.display_results() to show current results",
-        )
-        if not args.output_format == "prometheus":
-            logging.info(
-                "recorder.asn_counter and .prefix_counter dictionnaries have the full data"
-            )
-
-        # need to extract the namespace before we enter the thread,
-        # which runs with different locals
-        namespace = locals() | globals()
-        repl_thread = threading.Thread(
-            target=lambda: code.interact(local=namespace), daemon=True
-        )
-        repl_thread.start()
-        try:
-            repl_thread.join()
-        except KeyboardInterrupt:
-            pass
+        repl_thread(collector, namespace)
     else:
         collector()
 
-    if prom_server:
-        prom_server.shutdown()
-        prom_thread.join()
+    recorder.shutdown()
 
     if args.output_format != "null":
         # done collecting results, show a final stats dump
         recorder.display_results(args.output)
+    # for test suite
+    return recorder
+
+
+def main_wrap() -> None:  # pragma: nocover
+    """wrap main() to discard return value for executable wrapper
+
+    The script build by setuptools does sys.exit(main()) which is
+    incorrect here.
+    """
+    main()
 
 
 if __name__ == "__main__":  # pragma: nocover
-    main()
+    main_wrap()
